/**
  Ground Web Framework (c) 2011-2013 Optimal Bits Sweden AB
  MIT Licensed.
*/
/**
  Sequence Class

  This class represents a ordered collection of models.
  The sequence supports persistent storage, offline and
  automatic client<->server synchronization.

  Events:

*/

/// <reference path="container.ts" />
/// <reference path="../log.ts" />
/// <reference path="../using.ts" />
/// <reference path="../base.ts" />
/// <reference path="../model.ts" />
/// <reference path="../overload.ts" />
/// <reference path="../mutex.ts" />

module Gnd {

export interface ISeqModel {
  model: Model;
  id: string;
};

export interface SequenceEvents
{
  on(evt: string, listener: (...args: any[]) => void): Base;


  /**
  * Fired when any model in the collection changes any property.
  *
  * @event updated:
  * @param item {Model}
  * @param args {any}
  */
  on(evt: 'updated:', listener: (item: Model, args:any) => void): Base;

  /**
  * Fired when a model has been inserted in the sequence
  *
  * @event inserted:
  * @param item {Model}
  * @param index {Number}
  */
  on(evt: 'inserted:', listener: (item: Model, index: number) => void): Base;

  /**
  * Fired when a model has been removed from the sequence
  *
  * @event removed:
  * @param item {Model}
  * @param index {Number}
  */
  on(evt: 'removed:', listener: (item: Model, index: number) => void): Base;

  /**
  * Fired when the collection has been resynced.
  *
  * @event resynced:
  */
  on(evt: 'resynced:', listener: () => void): Base;
}

/**
 The {{#crossLink "Sequence"}}{{/crossLink}} class is used to represent an
 sorted set of models providing special method for typical sequence operations
 such as insert, move, push, etc

 Sequences are normally instantiated lazily, meaning that the sequence instance
 is created and afterwards the data is fetched asynchronously. The sequence
 can be used before the data has arrived, and as long as the code using the
 sequence is reactive (based on the events generated by the sequence)
 everything will work as expected.

 Sequences inherit the following methods from [Underscore / Lodash](http://http://underscorejs.org/#collections)

      forEach, each, map, reduce, reduceRight, find, detect, pluck,
      filter, select, reject, every, all, some, any, include,
      contains, invoke, max, min, sortBy, sortedIndex, toArray, size,
      first, rest, last, without, indexOf, lastIndexOf, isEmpty, groupBy


  @class Sequence
  @extends Container
  @constructor
  @param model {IModel}
  @param [opts] {ContainerOptions}
  @param [parent] {Model}
  @param [items] {ISeqModel[]}
 **/
export class Sequence extends Container implements SequenceEvents
{
  static mergeFns = {
    id: function(item){
      return item.id;
    },
    keyPath: function(item){
      return item.keyPath;
    },
    doc: function(item){
      return item.doc;
    },
    inSync: function(item){
      return item.insync;
    }
  };

  public updateFn: (args: any) => void;
  public deleteFn: (model: Model) => void;

  // Mutex
  private resyncMutex = Mutex();

  constructor(model: IModel, opts?: ContainerOptions, parent?: Model, items?: ISeqModel[])
  {
    super(model, opts, parent, items);

    var __this = this;

    var updateFn = function(args){
      __this.emit('updated:', this, args);
    };

    var deleteFn = (model)=>{
      for(var i = this.items.length-1; i >= 0; i--){
        if(this.items[i].model.id() === model.id()){
          this.remove(i, {nosync: true});
        }
      }
    };

    this._init(updateFn, deleteFn);

    this._promise = new Promise((resolve, reject) => {
      var keyPath = this.getKeyPath();
      if(keyPath && !this.opts.nosync){
        this.retain();

        using.storageQueue.all(keyPath).then((result) => {
          this.resync(result[0]);
          result[1].then((items?) => this.resync(items))
            .ensure(() => {
              resolve(this);
              this.release();
            }).fail((err) => {});
        });
      }else{
        resolve(this);
      }
    });
    this._promise.uncancellable = true;
  }

  private deleteItem(id: string, opts): Promise<any>
  {
    var idx = _.findIndex(this.items, {'id': id});

    if(idx === -1) return Promise.resolved(this); //already deleted
    return this.remove(idx, opts);
  }

  private insertBefore(refId: string, item: Model, opts?): Promise<any>
  {
    return this.insertItemBefore(refId, item, null, opts);
  }

  private insertItemBefore(refId: string, item: Model, id: string, opts): Promise<any>
  {
    var seqItem = {
      model: item,
      id: id,
      insync: !(_.isNull(id) || _.isUndefined(id))
    };

    opts = Util.extendClone(this.opts, opts);
    if(id) opts.id = id;

    var done = (id)=>{
      seqItem.id = id.id || seqItem.id;
      this.storageQueue.once('inserted:'+seqItem.id, (sid)=>{
        seqItem.insync = true;
      });
    }

    var index = _.findIndex(this.items, {'id': id});
    if(index !== -1){
      var next = this.items[index+1];
      if((!refId && !next) || refId === next.id){
        return Promise.resolved(); //already at the right place
      }else{
        return Promise.rejected(Error('Tried to insert duplicate container'));
      }
    }

    if(refId){
      index = _.findIndex(this.items, {'id': refId});
      if(index === -1){
        //refId not found perform a resync
        log('REFID not found. Resyncing');
        return this.triggerResync();
      }
    }else{
      //push last
      index = this.items.length;
    }

    //Handle the case when we insert an item from the server that have a pending item on this position
    while(opts.noremote && index > 0 && !this.items[index-1].insync){
      index--;
      refId = this.items[index].id;
    }

    this.initItems(item);
    this.items.splice(index, 0, seqItem);
    this.set('count', this.items.length);
    this._keepSynced && item.keepSynced();

    var promise;
    if(!opts || !opts.nosync){
      if(item.isPersisted() || item._persisting){
        promise = this.insertPersistedItemBefore(refId, item, opts).then<any>(done);
      }else{
        promise = item.save().then(()=>{
          return this.insertPersistedItemBefore(refId, item, opts).then(done);
        });
      }
    }else{
      promise = Promise.resolved();
    }

    this.emit('inserted:', item, index);
    return promise;
  }

  private insertPersistedItemBefore(id: string, item: Model, opts: {}): Promise<any>
  {
    var keyPath = this.getKeyPath();
    var itemKeyPath = item.getKeyPath();
    return this.storageQueue.insertBefore(keyPath, id, itemKeyPath, opts);
  }

  /**

    Pushes an item at the end of the sequence.

    @method push
    @param item {Model}
    @param [opts] {Any}
    @returns {Promise}
  **/
  push(item: Model, opts?): Promise<any>
  {
    return this.insertBefore(null, item, opts);
  }

  /**

    Inserts an item at the beginning of the sequence

    @method unshift
    @param item {Model}
    @param [opts] {Any}
    @returns {Promise}
  **/
  unshift(item: Model, opts?): Promise<any>
  {
    var firstId = this.items.length > 0 ? _.first(this.items).id : null;
    return this.insertBefore(firstId, item, opts);
  }

  /**

    Inserts an item at the given index position in the sequence.

    @method insert
    @param idx {Number}
    @param item {Model}
    @param [opts] {Any}
    @returns {Promise}
  **/
  insert(idx: number, item: Model, opts?): Promise<any>
  {
    var seqItem = this.items[idx];
    var id = seqItem ? seqItem.id : null;
    return this.insertBefore(id, item, opts);
  }

  /**

    Removes an item at the given index position in the sequence.

    @method remove
    @param idx {Number}
    @param [opts] {Any}
    @returns {Promise}
  **/
  remove(idx: number, opts?): Promise<any>
  {
    var promise;

    var item = this.items[idx];

    if(!item){
      return Promise.rejected(Error('index out of bounds'));
    }
    this.items.splice(idx, 1);
    this.deinitItems(item.model);
    this.set('count', this.items.length);

    opts = Util.extendClone(this.opts, opts);

    if(!opts || !opts.nosync){
      promise = this.storageQueue.deleteItem(this.getKeyPath(), item.id, opts);
    }else{
      promise = Promise.resolved();
    }

    this.emit('removed:', item.model, idx);
    return promise;
  }

  /**

    Moves an item from the given index position to another index position.

    @method move
    @param startIdx {Number}
    @param endIdx {Number}
    @param [opts] {Any}
    @returns {Promise}
  **/
  move(startIdx: number, endIdx: number, opts?): Promise<any>
  {
    var srcItem = this.items[startIdx];

    if(srcItem){
      endIdx = startIdx <= endIdx ? endIdx + 1 : endIdx;

      if(0 <= endIdx && endIdx <= this.items.length){
        var targetId = endIdx < this.items.length ? this.items[endIdx].id : null;

        srcItem.model.retain();
        return this.remove(startIdx).then(()=>{
          return this.insertBefore(targetId, srcItem.model, opts);
        });
      }
    }
    return Promise.rejected(Error("Invalid indexes:"+startIdx+", "+endIdx));
  }

  /**

    Gets an array with all the items stored in the sequence.

    @method getItems
    @returns {Model[]}
  **/
  public getItems(): Model[]
  {
    return _.pluck(this.items, 'model');
  }

  public startSync()
  {
    super.startSync();

    this.on('insertBefore:', (id, itemKeyPath, refId)=>{
      this.model.findById(itemKeyPath, true, {}).then((item)=>{
        this.insertItemBefore(refId, item, id, {noremote: true});
        item.release();
      });
    });

    this.on('deleteItem:', (id) => {
      this.deleteItem(id, {noremote: true});
    });
  }

  private execCmds(commands: MergeCommand[]): Promise<any>
  {
    var opts = {nosync: true};
    var item;
    return Gnd.Promise.map(commands, (cmd) => {
      switch(cmd.cmd) {
        case 'insertBefore':
          item = this.model.create(cmd.doc, this.autosync());
          item.autorelease();
          return this.insertItemBefore(cmd.refId, item, cmd.newId, opts);
        case 'removeItem':
          return this.deleteItem(cmd.id, opts);
        case 'update':
          item = this['find']((item) => cmd.doc._cid == item.id());
          item && item.resync(cmd.doc);
          return Promise.resolved();
        default:
          throw Error('Invalid command:'+cmd);
      }
    });
  }

  private triggerResync(): Promise<any>
  {
    var keyPath = this.getKeyPath();
    return this.storageQueue.all(keyPath, {}, {}).then((result) =>
      result[1].then((items) => this.resync(items))
    );
  }

  public resync(remoteItems?: any[]): Promise<any>
  {
    return this.resyncMutex(() => {
      if(remoteItems){
        return this._resync(remoteItems);
      }else{
        this.retain();
        using.storageQueue.allRemote(this.getKeyPath(), {}, {})
          .then((items) => this._resync(<any>items)).ensure(()=>this.release());
      }
    });
  }

  private _resync(items: any[]): Promise<any>
  {
    var commands = Sequence.merge(items, this.items, Sequence.mergeFns);
    return this.execCmds(commands).then(() => {
      this.emit('resynced:');
    });
  }

  static merge(source: any[], target: any[], fns: MergeFunctions): MergeCommand[]
  {
    source = source || [];
    var insertCommands: MergeCommand[] = [];
    var removeCommands: MergeCommand[] = [];
    var updateCommands: MergeCommand[] = [];
    var remainingItems = [];

    var sourceIds = _.map(source, function(item){
      return fns.id(item); //TODO: Change to id
    }).sort();

    //Determine which items to delete
    _.each(target, function(targetItem){
      if(fns.inSync(targetItem) && -1 === _.indexOf(sourceIds, fns.id(targetItem), true)){
        removeCommands.push({
          cmd: 'removeItem',
          id: fns.id(targetItem)
        });
      }else{
        remainingItems.push(targetItem);
      }
    });

    var i=0;
    var j=0;
    var targetItem, sourceItem;

    // insert new items on the right place
    while(i<remainingItems.length && j<source.length){
      targetItem = remainingItems[i];
      if(fns.inSync(targetItem)){
        sourceItem = source[j];
        if(fns.id(targetItem) === fns.id(sourceItem)){
          updateCommands.push({
            cmd: 'update',
            // id: fns.id(sourceItem),
            keyPath: fns.keyPath(sourceItem), //TODO: not always needed
            doc: fns.doc(sourceItem)
          });
          i++;
        }else{
          insertCommands.push({
            cmd: 'insertBefore',
            refId: fns.id(targetItem),
            newId: fns.id(sourceItem),
            keyPath: fns.keyPath(sourceItem), //TODO: not always needed
            doc: fns.doc(sourceItem)
          });
        }
        j++;
      }else{
        i++;
      }
    }

    //append remaining new items
    while(j<source.length){
      sourceItem = source[j];
      insertCommands.push({
        cmd: 'insertBefore',
        refId: null,
        newId: fns.id(sourceItem),
        keyPath: fns.keyPath(sourceItem), //TODO: see above
        doc: fns.doc(sourceItem)
      });
      j++;
    }

    //remove remaining old items
    while(i<remainingItems.length){
      targetItem = remainingItems[i];
      if(fns.inSync(targetItem)){
        removeCommands.push({
          cmd: 'removeItem',
          id: fns.id(targetItem)
        });
      }
      i++;
    }

    // Adapt insert commands if pointing at deleted item
    _.each(insertCommands, function(insertCmd){
      var found = _.find(removeCommands, function(removeCmd: any){
        return removeCmd.id === insertCmd.refId;
      });
      if(found){
        insertCmd.refId = null;
      }
    });


    // return the sequence of commands that transforms the target sequence according
    // to the source
    // it is important that the removecommands come before the insertcommands
    return removeCommands.concat(insertCommands).concat(updateCommands);
  }
}

//
// Underscore methods that we want to implement on the Sequence
//
var methods =
  ['each', 'map', 'reduce', 'reduceRight', 'find', 'detect', 'pluck',
    'filter', 'select', 'reject', 'every', 'all', 'some', 'any', 'include',
    'contains', 'invoke', 'max', 'min', 'toArray', 'size',
    'first', 'rest', 'last', 'without', 'indexOf', 'lastIndexOf', 'isEmpty']

// Mix in each Underscore method as a proxy to `Sequence#items`.
// The pluck is a candidate for optimization
_.each(methods, function(method) {
  Sequence.prototype[method] = function() {
    return _[method].apply(_, [_.pluck(this.items, 'model')].concat(_.toArray(arguments)))
  }
});

export interface MergeFunctions {
  id: (item: {}) => string;
  keyPath: (item: {}) => string[];
  doc: (item: {}) => {};
  inSync: (item: {}) => boolean;
}

export interface MergeCommand {
  cmd: string;
  id?: string;
  refId?: string;
  newId?: string;
  keyPath?: string[];
  doc?: {};
}

}
